using SecretLabs.NETMF.Hardware;
using SecretLabs.NETMF.Hardware.Netduino;

namespace NetQuadCopter
{
    public class ENGManager
    {
        /// <summary>
        /// Puertos para controlar los motores
        /// </summary>
        private PWM FrontMotor;
        private PWM RearMotor;
        private PWM LeftMotor;
        private PWM RightMotor;

        /// <summary>
        /// Para debug almaceno los ltimos valores establecidos
        /// </summary>
        private short currentFront;
        private short currentRear;
        private short currentLeft;
        private short currentRight;

        /// <summary>
        /// Constructor
        /// </summary>
        public ENGManager()
        {
            FrontMotor = new SecretLabs.NETMF.Hardware.PWM(Pins.GPIO_PIN_D5);
            RearMotor =  new SecretLabs.NETMF.Hardware.PWM(Pins.GPIO_PIN_D6);
            LeftMotor =  new SecretLabs.NETMF.Hardware.PWM(Pins.GPIO_PIN_D9);
            RightMotor = new SecretLabs.NETMF.Hardware.PWM(Pins.GPIO_PIN_D10);

            // Inicialmente apagamos los motores
            SwitchOff();
        }

        /// <summary>
        /// Apaga los motores
        /// </summary>
        public void SwitchOff()
        {
            // Apagamos los motores
            FrontMotor.SetDutyCycle(0);
            RearMotor.SetDutyCycle(0);
            LeftMotor.SetDutyCycle(0);
            RightMotor.SetDutyCycle(0);
        }

        /// <summary>
        /// Devuelve una traza con valores de variables privadas. Usado para depurar
        /// </summary>
        /// <returns></returns>
        public string GetDebug()
        {
            return "Front: " + currentFront + ", Rear: " + currentRear + ", Left: " + currentLeft + ", Right: " + currentRight;
        }


        /// <summary>
        /// Mtodo pblico que se habr de llamar periodicamente para corregir la potencia
        /// de cada motor y as corregir la direccin
        /// </summary>
        /// <param name="pitch">Cabeceo o profundidad, indicada en radianes PI<->-PI</param>
        /// <param name="roll">Alabeo, indicada en radianes PI<->-PI</param>
        /// <param name="yaw">Direccin, indicada en radianes PI<->-PI</param>
        /// <param name="enginePower">Potencia general de los motores en porcentaje. 50% indica que se debe quedar a la misma altura.</param>
        /// <param name="xAxis">Porcentaje del angulo de giro sobre el eje X. 50% indica que no gira</param>
        /// <param name="yAxis">Porcentaje del angulo de giro sobre el eje Y. 50% indica que no gira</param>
        /// <param name="zAxis">Porcentaje del angulo de giro sobre el eje Z. 50% indica que no gira</param>
        public void SetEngines(
            float pitch,
            float roll,
            float yaw,
            short enginePower,
            short xAxis,
            short yAxis,
            short zAxis)
        {
            // Calculamos el valor (en %) de cada motor

            // Inicialmente ponemos la potencia indicada por la radio
            short front = enginePower;
            short rear = enginePower; 
            short left = enginePower;
            short right = enginePower;

            // Pitch
            short p = (short)System.Math.Round(pitch * 100 / NetQuadCopter.Math.pio2);
            front = (short)(front - p);
            rear = (short)(rear + p);

            // Rool
            short r = (short)System.Math.Round(roll * 100 / NetQuadCopter.Math.pio2);
            right = (short)(right + r);
            left = (short)(left - r);

            // Yaw
            // TO-DO: -> Me hara falta una brjula electrnica. Esto de momento no lo gestiono

            // Por si nos pasamos...
            if (front < 0) front = 0;
            if (rear < 0) rear = 0;
            if (left < 0) left = 0;
            if (right < 0) right = 0;

            if (front > 100) front = 100;
            if (rear > 100) rear = 100;
            if (left > 100) left = 100;
            if (right > 100) right = 100;

            // Variamos los valores calculados anteriormente para llevar el helicpero a su posicin estable
            // con los valores indicados por la emisora
            left += (short)(xAxis - 50);
            right -= (short)(xAxis - 50);
            front -= (short)(yAxis - 50);
            rear += (short)(yAxis - 50);

            // Parte del Yaw, eje Z (esto depender de para que lado gire cada motor)
            left += (short)(zAxis - 50);
            right -= (short)(zAxis - 50);
            front -= (short)(zAxis - 50);
            rear += (short)(zAxis - 50);

            // Aplicamos la potencia a los motores
            FrontMotor.SetDutyCycle((uint)front);
            RearMotor.SetDutyCycle((uint)rear);
            LeftMotor.SetDutyCycle((uint)left);
            RightMotor.SetDutyCycle((uint)right);

            // Guardo los valores por si quiero depurar y consultarlos
            currentFront = front;
            currentRear = rear;
            currentLeft = left;
            currentRight = right;
        }
    }
}
