﻿using System;
using System.Threading;
using Microsoft.SPOT;
using Microsoft.SPOT.Hardware;
using SecretLabs.NETMF.Hardware.Netduino;

namespace NetQuadCopter
{
    public class Program
    {
        // Timer que se encargará de actualizar la potencia de cada salida PWM cada cierto tiempo
        private static Timer loopTimer;

        // LED de la placa (indicará si está encendido o no)
        private static OutputPort internalLED = null;

        // Menejadores de la radio, la IMU y los motores. Los tres grandes pilares
        private static RCManager rcManager;
        private static IMUManager imuManager;
        private static ENGManager engManager;

        // Para controlar cuando llamar al recolector de basura de forma explícita
        private static int numLoops = 0;

        // Para debug
        private static int debugCounter = 0;

        public static void Main()
        {
            try
            {
                // intervalo en milisegundos entre actualizaciones del cuatricóptero
                // Arducopter creo que trabaja a 200 Hz lo cual significaría un intervalo de 5ms
                // lo cual veo muchísimo para Netduino
                byte hz = 10; // Hz

                byte interval = (byte)(1000 / hz);

                // Memoria
                //Debug.EnableGCMessages(true); // -> Al principio del programa para activar la información de depuración del GC

                // Inicializo el LED interno de la placa y lo apago
                internalLED = new OutputPort(Pins.ONBOARD_LED, false);

                // Creo los manejadores
                rcManager = new RCManager(false, true);
                imuManager = new IMUManager(interval);
                engManager = new ENGManager();

                // Inicializo el timer
                loopTimer = new Timer(new System.Threading.TimerCallback(loopTimerCallback), null, 0, interval);

                // Thread que controlará el parpadeo del LED de la placa
                Thread blinkLedThread = new Thread(new ThreadStart(blinkLed));
                blinkLedThread.Start();

                // Espera infinita para evitar la terminación del programa
                Thread.Sleep(Timeout.Infinite);
            }
            catch (Exception ex)
            {
                // Paramos los motores
                engManager.SwitchOff();

                // REBOOT
                PowerState.RebootDevice(false);
                //throw;
            }
        }

        /// <summary>
        /// Método que se encarga de hacer parpadear el LED de la placa.
        /// - Parpadeo lento -> Funcionando
        /// - Parpadeo rápido -> no recibe señal de la radio
        /// </summary>
        private static void blinkLed()
        {
            while (true)
            {
                if (rcManager.PowerChannel)
                {
                    // Enciendo el LED de la placa
                    internalLED.Write(true);
                    Thread.Sleep(800);

                    // Apago el LED de la placa
                    internalLED.Write(false);
                    Thread.Sleep(800);
                }
                else
                {
                    // Enciendo el LED de la placa
                    internalLED.Write(true);
                    Thread.Sleep(200);

                    // Apago el LED de la placa
                    internalLED.Write(false);
                    Thread.Sleep(200);
                }
            }
        }

        /// <summary>
        /// Delegado del timer que se encarga de actualizar la potencia de los motores
        /// </summary>
        /// <param name="obj"></param>
        private static void loopTimerCallback(object obj)
        {
            try
            {
                //Debug.GC(true); // -> Al principio de una rutina (creo que dice la memoria libre)

                // Solo ejecutamos el código si está activo el canal de "Power" de la radio
                if (rcManager.PowerChannel)
                {
                    // Hago la llamada al controlador de motores pasandole la actual potencia de los motores indicados por la radio
                    engManager.SetEngines(imuManager.Pitch, imuManager.Roll, imuManager.Yaw, rcManager.EngineChannel, rcManager.XAxisChannel, rcManager.YAxisChannel, rcManager.ZAxisChannel);

                    // Debug de los motores
                    Debug.Print(debugCounter + " - " + engManager.GetDebug());

                    //Microsoft.SPOT.Debug.GC(false).ToString()

                    // Debug de la IMU
                    //Debug.Print(debugCounter + " - " + imuManager.GetDebug());

                    // Debug de la emisora
                    //Debug.Print(debugCounter + " - " + rcManager.GetDebug(0));
                }
                else
                {
                    // Paramos los motores
                    engManager.SwitchOff();

                    Debug.Print(debugCounter + " - " + "Apagado...");
                }

                // Recolector de basura cada 50 iteracciones
                numLoops++;
                if (numLoops == 50)
                {
                    GC.WaitForPendingFinalizers();
                    numLoops = 0;
                }

                // Para debug
                debugCounter++;
            }
            catch (Exception ex)
            {
                // Paramos los motores
                engManager.SwitchOff();

                throw;
            }
        }
    }
}
