using System;
using Microsoft.SPOT;
using Microsoft.SPOT.Hardware;
using NetQuadCopter.libraries;
using SecretLabs.NETMF.Hardware.Netduino;

namespace NetQuadCopter
{
    /// <summary>
    /// Estructura para guardar la configuracin y los mrgenes de calibracin de cada canal
    /// </summary>
    struct ChannelConfig
    {
        public Cpu.Pin pin;
        public bool invert;
        public bool calibrated;

        private byte _id;

        private short _value;
        public short value
        {
            get {
                // Si an no se ha calibrado devolvemos 50%
                if (!calibrated) return 50;
                else return _value;
            }
            set {
                _value = value;

                // Compruebo si he superado los lmites para modificarlos
                if (value < minValue)
                {
                    // He recibido un valor mnimo menor que el que tena al principio
                    minValue = value;
                    Debug.Print("Establecido el valor mnimo del canal " + _id + " a:" + value);
                }
                if (value > maxValue)
                {
                    // He recibido un valor mximo mayor que el que tena al principio
                    maxValue = value;
                    Debug.Print("Establecido el valor mximo del canal " + _id + " a:" + value);
                }
            }
        }


        // Uso estas variables para calibrar el canal
        public float topMarginRate;
        public float bottomMarginRate;

        private short _minValue;
        public short minValue
        {
            get { return _minValue; }
            set {
                _minValue = value;
                _updateMargins();
            }
        }

        private short _neutralValue;
        public short neutralValue
        {
            get { return _neutralValue; }
            set
            {
                _neutralValue = value;
                _updateMargins();
            }
        }

        private short _maxValue;
        public short maxValue
        {
            get { return _maxValue; }
            set
            {
                _maxValue = value;
                _updateMargins();
            }
        }

        private void _updateMargins()
        {
            try
            {
                topMarginRate = 50.0f / (maxValue - neutralValue);
                bottomMarginRate = 50.0f / (neutralValue - minValue);
            }
            catch (Exception ex)
            {
                throw;
            }
        }

        /// <summary>
        /// Constructor que configura el canal para auto-calibracin
        /// </summary>
        /// <param name="pPin">Pin en el que est conectado</param>
        /// <param name="pInvert">Indica si hay que invertir las mediciones</param>
        public ChannelConfig(byte id, Cpu.Pin pPin, bool pInvert)
        {
            this._id = id;
            this._value = 0;
            this._minValue = 0;
            this._neutralValue = 0;
            this._maxValue = 100;
            this.invert = pInvert;
            this.topMarginRate = 0;
            this.bottomMarginRate = 0;
            this.pin = pPin;
            this.calibrated = false;
        }

        /// <summary>
        /// Constructor que configura el canal con unos valores prefijados
        /// </summary>
        /// <param name="id">Id del canal</param>
        /// <param name="pPin">Pin en el que est conectado</param>
        /// <param name="pInvert">Indica si hay que invertir las mediciones</param>
        /// <param name="pMinValue">Valor mnimo que puede tomar el canal</param>
        /// <param name="pMaxValue">Valor mximo que puede tomar el canal</param>
        /// <param name="pNeutralValue">Valor neutral que tiene el canal</param>
        public ChannelConfig(byte id, Cpu.Pin pPin, bool pInvert, short pMinValue, short pMaxValue, short pNeutralValue)
        {
            this._id = id;
            this._minValue = pMinValue;
            this._neutralValue = pNeutralValue;
            this._value = _neutralValue;
            this._maxValue = pMaxValue;
            this.invert = pInvert;
            this.topMarginRate = 50.0f / (pMaxValue - pNeutralValue);
            this.bottomMarginRate = 50.0f / (pNeutralValue - pMinValue);
            this.pin = pPin;
            this.calibrated = true;
        }
    }

    /// <summary>
    /// Clase principal encargada de manejar todos los canales de la emisora
    /// </summary>
    public class RCManager
    {
        private const byte XAxis        = 0;
        private const byte YAxis        = 1;
        private const byte ZAxis        = 2;
        private const byte EngineAxis   = 3;
        private const byte PowerAxis    = 4;
        private const byte Other        = 5;

        // Indica si se han de normalizar los canales
        private bool normalizeChannels;

        /// <summary>
        /// Receptor de radio
        /// </summary>
        private RXReader reader;

        /// <summary>
        /// Array para almacenar el valor de cada canal
        /// </summary>
        private ChannelConfig[] channels = new ChannelConfig[6];

        #region Propiedades pblicas

        /// <summary>
        /// Calcula el valor actual de canal controlando sus mrgenes de calibracin
        /// </summary>
        /// <param name="channel">Configuracin del canal</param>
        /// <returns>Valor, en porcentaje del canal</returns>
        private short CalculateChannel(byte channel)//ChannelConfig channel)
        {
            try
            {
                if (!channels[channel].calibrated) return 50;

                // Invierto si toca el valor
                short result = channels[channel].value;

                // Y ahora lo extrapolo teniendo en cuenta los mrgenes de calibrado
                if (result == channels[channel].neutralValue) return 50;

                if (result > channels[channel].neutralValue)
                    result = (short)System.Math.Round(50 + (result - channels[channel].neutralValue) * channels[channel].topMarginRate);
                else
                    result = (short)System.Math.Round((result - channels[channel].minValue) * channels[channel].bottomMarginRate);

                // Si hay que invertir...
                if (channels[channel].invert) result = (short)(100 - result);

                // Normalizamos para que la escala sea logartmica
                if (normalizeChannels) result = normalizeChannel(result);

                return result;
            }
            catch (Exception ex)
            {
                throw;
            }
        }

        public short XAxisChannel{
            get { return CalculateChannel(XAxis); }
        }
        public short YAxisChannel{
            get { return CalculateChannel(YAxis); }
        }
        public short ZAxisChannel{
            get { return CalculateChannel(ZAxis); }
        }
        public short EngineChannel{
            get { return CalculateChannel(EngineAxis); }
        }
        public bool PowerChannel{
            get { return CalculateChannel(PowerAxis) > 60; }
        }
        public short OtherChannel{
            get { return CalculateChannel(Other); }
        }

        #endregion

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="autoconfig">Indica si se ha de autocalibrar o por el contrario usa los valores que tiene en el cdigo
        /// En teora habra que hacer pruebas para que se autocalibre y apuntar esos valores para posteriormente ponerlos
        /// en el cdigo y hacer que arranque con estos valores</param>
        /// <param name="normalizeChannels">Indica si los canales hay que normalizarlos con una escala logartmica</param>
        public RCManager(bool autoconfig, bool normalizeChannels)
        {
            try
            {
                // normalizar canales?
                this.normalizeChannels = normalizeChannels;

                // Almaceno los canales en un array para facilitar su acceso
                // Calibracin automtica
                if (autoconfig)
                {
                    channels[XAxis] = new ChannelConfig(XAxis, Pins.GPIO_PIN_D0, false); // XAxis (canal 1)
                    channels[YAxis] = new ChannelConfig(YAxis, Pins.GPIO_PIN_D1, true);  // YAxis (canal 2)
                    channels[ZAxis] = new ChannelConfig(ZAxis, Pins.GPIO_PIN_D3, true);  // ZAxis (canal 4)
                    channels[EngineAxis] = new ChannelConfig(EngineAxis, Pins.GPIO_PIN_D2, false); // EngineAxis (canal 3)
                    channels[PowerAxis] = new ChannelConfig(PowerAxis, Pins.GPIO_PIN_D4, false); // PowerAxis (canal 5)
                    channels[Other] = new ChannelConfig(Other, Pins.GPIO_PIN_D7, false); // Other (canal 6)
                }
                else
                {
                    // Calibracin manual -> Poner aqu los valores pertinentes
                    channels[XAxis] = new ChannelConfig(XAxis, Pins.GPIO_PIN_D0, false, 15, 100, 62); // XAxis
                    channels[YAxis] = new ChannelConfig(YAxis, Pins.GPIO_PIN_D1, true, 0, 100, 47); // YAxis
                    channels[ZAxis] = new ChannelConfig(ZAxis, Pins.GPIO_PIN_D3, true, 0, 98, 49); // ZAxis
                    channels[EngineAxis] = new ChannelConfig(EngineAxis, Pins.GPIO_PIN_D2, false, 0, 98, 49); // EngineAxis
                    channels[PowerAxis] = new ChannelConfig(PowerAxis, Pins.GPIO_PIN_D4, false, 0, 100, 40); // PowerAxis
                    channels[Other] = new ChannelConfig(Other, Pins.GPIO_PIN_D7, false, 0, 96, 40); // Other
                }

                // Inicializo el receptor para que reciba la radio y actualize los valores en memoria
                reader = new RXReader(6);
                reader.addChannel(channels[XAxis].pin,      XAxis);
                reader.addChannel(channels[YAxis].pin,      YAxis);
                reader.addChannel(channels[ZAxis].pin,      ZAxis);
                reader.addChannel(channels[EngineAxis].pin, EngineAxis);
                reader.addChannel(channels[PowerAxis].pin,  PowerAxis);
                reader.addChannel(channels[Other].pin,      Other);

                // Me suscribo al evento
                reader.channelUpdated += new RXReader.ChannelUpdated(reader_channelUpdated);
            }
            catch (Exception ex)
            {
                throw;
            }
        }


        /// <summary>
        /// Usado para poder depurar
        /// </summary>
        /// <param name="channel"></param>
        /// <returns></returns>
        public string GetDebug(byte channel)
        {
            return "Channel " + channel + ": (Min:" + channels[channel].minValue + "|Neu:" + channels[channel].neutralValue + "|Max:" + channels[channel].maxValue + "|Val:" + channels[channel].value + ") -> " + CalculateChannel(channel);
        }

        /// <summary>
        /// Establece los valores actuales de cada canal como valor neutral
        /// </summary>
        public void SetNeutralValues()
        {
            try
            {
                for(int i = 0 ; i < channels.Length ; i++)
                {
                    channels[i].neutralValue = channels[i].value;
                    
                    Debug.Print("Establecido el valor neutral del canal " + i + " a:" + channels[i].neutralValue);
                }
            }
            catch (Exception ex)
            {
                throw;
            }
        }

        /// <summary>
        /// Evento de nueva recepcin por un canal de la emisora
        /// </summary>
        /// <param name="channel">Informacin del canal</param>
        /// <param name="oldValue">ltimo valor recibido</param>
        /// <param name="newValue">Nuevo valor recibido</param>
        private void reader_channelUpdated(ChannelInfo channel, float oldValue, float newValue)
        {
            try
            {
                // Si no se ha calibrado y recibo un 0 es porque no estoy recibiendo de la emisora al arrancar
                if (!channels[channel.channelID].calibrated && (newValue == 0 || newValue == 1))
                {
                    Debug.Print("Valor recibido no vlido para el canal " + channel.channelID + ": " + oldValue + "," + newValue);
                    return;
                }

                // Lo paso a porcentaje
                short nVal = (short)System.Math.Round(newValue * 100);

                if (!channels[channel.channelID].calibrated)
                {
                    channels[channel.channelID].neutralValue = nVal;
                    
                    if (nVal > 50)
                    {
                        channels[channel.channelID].neutralValue = (short)System.Math.Min((int)nVal, 60);

                        channels[channel.channelID].minValue += (short)((nVal - 50) * 2);
                        //channels[channel.channelID].minValue = (short)System.Math.Max((int)channels[channel.channelID].minValue, 40);
                    }
                    if (nVal < 50)
                    {
                        channels[channel.channelID].neutralValue = (short)System.Math.Max((int)nVal, 40);

                        channels[channel.channelID].maxValue -= (short)((50 - nVal) * 2);
                        //channels[channel.channelID].minValue = (short)System.Math.Min((int)channels[channel.channelID].maxValue, 60);
                    }

                    channels[channel.channelID].calibrated = true;
                    Debug.Print("Canal " + channel.channelID + " calibrado: " + channels[channel.channelID].minValue + "," + channels[channel.channelID].neutralValue + "," + channels[channel.channelID].maxValue);
                }

                // Actualizo el valor
                channels[channel.channelID].value = nVal;
            }
            catch (Exception ex)
            {
                throw;
            }
        }



        /// <summary>
        /// Modifica el valor obtenido por la emisora de manera que sea una escala logartmica
        /// y as conseguir tener ms "tacto" con los mandos
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        private short normalizeChannel(short value)
        {
            short result = 0;

            if (value < 50)
                result = (short)(value * (100 - value) / 50);
            else if (value == 50)
                result = 50;
            else
                // value > 50
                result = (short)((5000 + value * (value - 100)) / 50);

            // Comprobaciones por una posible desviacin con los decimales
            if (result < 0) return 0;
            if (result > 100) return 100;
            else return (short)result;
        }
    }
}
